import 'package:equatable/equatable.dart';
import 'package:flutter/material.dart';
import 'package:hydrated_bloc/hydrated_bloc.dart';

import 'package:minehunter/models/game/cell.dart';
import 'package:minehunter/models/game/game.dart';
import 'package:minehunter/models/settings/settings_game.dart';
import 'package:minehunter/models/settings/settings_global.dart';
import 'package:minehunter/models/types.dart';
import 'package:minehunter/utils/tools.dart';

part 'game_state.dart';

class GameCubit extends HydratedCubit<GameState> {
  GameCubit()
      : super(GameState(
          currentGame: Game.createEmpty(),
        ));

  void updateState(Game game) {
    emit(GameState(
      currentGame: game,
    ));
  }

  void refresh() {
    final Game game = Game(
      // Settings
      gameSettings: state.currentGame.gameSettings,
      globalSettings: state.currentGame.globalSettings,
      // State
      isRunning: state.currentGame.isRunning,
      isStarted: state.currentGame.isStarted,
      isFinished: state.currentGame.isFinished,
      animationInProgress: state.currentGame.animationInProgress,
      // Base data
      board: state.currentGame.board,
      sizeHorizontal: state.currentGame.sizeHorizontal,
      sizeVertical: state.currentGame.sizeVertical,
      isBoardMined: state.currentGame.isBoardMined,
      // Game data
      minesCount: state.currentGame.minesCount,
      reportMode: state.currentGame.reportMode,
      gameWin: state.currentGame.gameWin,
      gameFail: state.currentGame.gameFail,
    );
    // game.dump();

    updateState(game);
  }

  void startNewGame({
    required GameSettings gameSettings,
    required GlobalSettings globalSettings,
  }) {
    final Game newGame = Game.createNew(
      // Settings
      gameSettings: gameSettings,
      globalSettings: globalSettings,
    );

    newGame.dump();

    updateState(newGame);
    refresh();
  }

  void quitGame() {
    state.currentGame.isRunning = false;
    refresh();
  }

  void resumeSavedGame() {
    state.currentGame.isRunning = true;
    refresh();
  }

  void deleteSavedGame() {
    state.currentGame.isRunning = false;
    state.currentGame.isFinished = true;
    refresh();
  }

  void updateBoard(Board board) {
    state.currentGame.board = board;
    refresh();
  }

  void addMines({
    required int forbiddenRow,
    required int forbiddenCol,
  }) {
    state.currentGame.isStarted = true;
    state.currentGame.addMines(
      forbiddenRow: forbiddenRow,
      forbiddenCol: forbiddenCol,
    );
    state.currentGame.isBoardMined = true;
    refresh();
  }

  void reportCell({
    required int row,
    required int col,
  }) {
    if (!state.currentGame.board[row][col].isExplored) {
      toggleCellMark(row, col);
    }
  }

  void walkOnCell({
    required int row,
    required int col,
  }) {
    setCellAsExplored(row, col);

    if (state.currentGame.board[row][col].minesCountAround == 0) {
      final List<List<int>> safeCells = getAllSafeCellsAround(row, col);
      for (int safeCellIndex = 0; safeCellIndex < safeCells.length; safeCellIndex++) {
        final int safeCellRow = safeCells[safeCellIndex][0];
        final int safeCellCol = safeCells[safeCellIndex][1];
        if (!state.currentGame.board[safeCellRow][safeCellCol].isExplored) {
          walkOnCell(
            row: safeCellRow,
            col: safeCellCol,
          );
        }
      }
    }
  }

  List<List<int>> getAllSafeCellsAround(int row, int col) {
    final List<List<Cell>> board = state.currentGame.board;
    final int sizeHorizontal = board.length;
    final int sizeVertical = board[0].length;

    final List<List<int>> safeCellsCoordinates = [];

    if (board[row][col].minesCountAround == 0) {
      for (int deltaRow = -1; deltaRow <= 1; deltaRow++) {
        for (int deltaCol = -1; deltaCol <= 1; deltaCol++) {
          final int candidateRow = row + deltaRow;
          final int candidateCol = col + deltaCol;
          if ((candidateRow >= 0 && candidateRow < sizeVertical) &&
              (candidateCol >= 0 && candidateCol < sizeHorizontal) &&
              !board[candidateRow][candidateCol].isExplored) {
            safeCellsCoordinates.add([candidateRow, candidateCol]);
          }
        }
      }
    }

    return safeCellsCoordinates;
  }

  void setCellAsExplored(int row, int col) {
    state.currentGame.board[row][col].isExplored = true;
    state.currentGame.board[row][col].isMarked = false;

    // Boom?
    if (state.currentGame.board[row][col].isMined) {
      // Boom!
      state.currentGame.board[row][col].isExploded = true;
    }

    refresh();
  }

  void toggleCellMark(int row, int col) {
    state.currentGame.board[row][col].isMarked = !state.currentGame.board[row][col].isMarked;
    refresh();
  }

  void updateReportMode(bool reportMode) {
    state.currentGame.reportMode = reportMode;
    refresh();
  }

  bool checkGameIsFinished() {
    final Game currentGame = state.currentGame;

    final Board board = currentGame.board;
    final int sizeHorizontal = board.length;
    final int sizeVertical = board[0].length;

    currentGame.printGrid();

    updateGameWin(false);
    updateGameFail(false);

    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        // Walked on a mine
        if (board[row][col].isExploded == true) {
          updateGameFail(true);
          return true;
        }
      }
    }

    for (int row = 0; row < sizeVertical; row++) {
      for (int col = 0; col < sizeHorizontal; col++) {
        if (
            // Mine not already found
            (board[row][col].isMined == true && board[row][col].isMarked == false) ||
                // Safe cell marked as mined
                (board[row][col].isMined == false && board[row][col].isMarked == true)) {
          return false;
        }
      }
    }

    printlog('-> ok all mines found!');
    updateGameWin(true);

    return true;
  }

  void updateGameWin(bool value) {
    state.currentGame.gameWin = value;
    if (true == value) {
      state.currentGame.isFinished = true;
    }
    refresh();
  }

  void updateGameFail(bool value) {
    state.currentGame.gameFail = value;
    if (true == value) {
      state.currentGame.isFinished = true;
    }
    refresh();
  }

  void updateAnimationInProgress(bool animationInProgress) {
    state.currentGame.animationInProgress = animationInProgress;
    refresh();
  }

  void setAnimatedBackground(List animatedCellsPattern) {
    for (int row = 0; row < state.currentGame.sizeVertical; row++) {
      for (int col = 0; col < state.currentGame.sizeHorizontal; col++) {
        state.currentGame.board[row][col].isAnimated = animatedCellsPattern[row][col];
      }
    }
    refresh();
  }

  void resetAnimatedBackground() {
    for (int row = 0; row < state.currentGame.sizeVertical; row++) {
      for (int col = 0; col < state.currentGame.sizeHorizontal; col++) {
        state.currentGame.board[row][col].isAnimated = false;
      }
    }
  }

  @override
  GameState? fromJson(Map<String, dynamic> json) {
    final Game currentGame = json['currentGame'] as Game;

    return GameState(
      currentGame: currentGame,
    );
  }

  @override
  Map<String, dynamic>? toJson(GameState state) {
    return <String, dynamic>{
      'currentGame': state.currentGame.toJson(),
    };
  }
}
